LIBRARY "d3dx8d.dll"
EXPORTS
D3DXAssembleShader@24
D3DXAssembleShaderFromFileA@20
D3DXAssembleShaderFromFileW@20
D3DXBoxBoundProbe@16
D3DXCheckCubeTextureRequirements@24
D3DXCheckTextureRequirements@28
D3DXCheckVolumeTextureRequirements@32
D3DXCleanMesh@12
D3DXColorAdjustContrast@12
D3DXColorAdjustSaturation@12
D3DXCompileEffect@16
D3DXCompileEffectFromFileA@12
D3DXCompileEffectFromFileW@12
D3DXComputeBoundingBox@20
D3DXComputeBoundingSphere@20
D3DXComputeNormals@4
D3DXCreateBox@24
D3DXCreateBuffer@8
D3DXCreateCubeTexture@28
D3DXCreateCubeTextureFromFileA@12
D3DXCreateCubeTextureFromFileExA@52
D3DXCreateCubeTextureFromFileExW@52
D3DXCreateCubeTextureFromFileInMemory@16
D3DXCreateCubeTextureFromFileInMemoryEx@56
D3DXCreateCubeTextureFromFileW@12
D3DXCreateCylinder@32
D3DXCreateEffect@20
D3DXCreateFont@12
D3DXCreateFontIndirect@12
D3DXCreateMatrixStack@8
D3DXCreateMesh@24
D3DXCreateMeshFVF@24
D3DXCreatePMeshFromStream@24
D3DXCreatePolygon@20
D3DXCreateRenderToEnvMap@24
D3DXCreateRenderToSurface@28
D3DXCreateSPMesh@20
D3DXCreateSkinMesh@28
D3DXCreateSkinMeshFVF@28
D3DXCreateSkinMeshFromMesh@12
D3DXCreateSphere@24
D3DXCreateSprite@8
D3DXCreateTeapot@12
D3DXCreateTextA@28
D3DXCreateTextW@28
D3DXCreateTexture@32
D3DXCreateTextureFromFileA@12
D3DXCreateTextureFromFileExA@56
D3DXCreateTextureFromFileExW@56
D3DXCreateTextureFromFileInMemory@16
D3DXCreateTextureFromFileInMemoryEx@60
D3DXCreateTextureFromFileW@12
D3DXCreateTextureFromResourceA@16
D3DXCreateTextureFromResourceExA@60
D3DXCreateTextureFromResourceExW@60
D3DXCreateTextureFromResourceW@16
D3DXCreateTorus@28
D3DXCreateVolumeTexture@36
D3DXDeclaratorFromFVF@8
D3DXFVFFromDeclarator@8
D3DXFilterCubeTexture@16
D3DXFilterTexture@16
D3DXFilterVolumeTexture@16
D3DXGeneratePMesh@28
D3DXGetErrorStringA@12
D3DXGetErrorStringW@12
D3DXGetFVFVertexSize@4
D3DXIntersect@32
D3DXLoadMeshFromX@28
D3DXLoadMeshFromXof@28
D3DXLoadSkinMeshFromXof@36
D3DXLoadSurfaceFromFileA@32
D3DXLoadSurfaceFromFileInMemory@36
D3DXLoadSurfaceFromFileW@32
D3DXLoadSurfaceFromMemory@40
D3DXLoadSurfaceFromResourceA@36
D3DXLoadSurfaceFromResourceW@36
D3DXLoadSurfaceFromSurface@32
D3DXLoadVolumeFromMemory@44
D3DXLoadVolumeFromVolume@32
D3DXMatrixAffineTransformation@20
D3DXMatrixInverse@12
D3DXMatrixLookAtLH@16
D3DXMatrixLookAtRH@16
D3DXMatrixMultiply@12
D3DXMatrixOrthoLH@20
D3DXMatrixOrthoOffCenterLH@28
D3DXMatrixOrthoOffCenterRH@28
D3DXMatrixOrthoRH@20
D3DXMatrixPerspectiveFovLH@20
D3DXMatrixPerspectiveFovRH@20
D3DXMatrixPerspectiveLH@20
D3DXMatrixPerspectiveOffCenterLH@28
D3DXMatrixPerspectiveOffCenterRH@28
D3DXMatrixPerspectiveRH@20
D3DXMatrixReflect@8
D3DXMatrixRotationAxis@12
D3DXMatrixRotationQuaternion@8
D3DXMatrixRotationX@8
D3DXMatrixRotationY@8
D3DXMatrixRotationYawPitchRoll@16
D3DXMatrixRotationZ@8
D3DXMatrixScaling@16
D3DXMatrixShadow@12
D3DXMatrixTransformation@28
D3DXMatrixTranslation@16
D3DXMatrixTranspose@8
D3DXMatrixfDeterminant@4
D3DXPlaneFromPointNormal@12
D3DXPlaneFromPoints@16
D3DXPlaneIntersectLine@16
D3DXPlaneNormalize@8
D3DXPlaneTransform@12
D3DXQuaternionBaryCentric@24
D3DXQuaternionExp@8
D3DXQuaternionInverse@8
D3DXQuaternionLn@8
D3DXQuaternionMultiply@12
D3DXQuaternionNormalize@8
D3DXQuaternionRotationAxis@12
D3DXQuaternionRotationMatrix@8
D3DXQuaternionRotationYawPitchRoll@16
D3DXQuaternionSlerp@16
D3DXQuaternionSquad@24
D3DXQuaternionToAxisAngle@12
D3DXSaveMeshToX@24
D3DXSimplifyMesh@28
D3DXSphereBoundProbe@16
D3DXTesselateMesh@20
D3DXValidMesh@8
D3DXVec2BaryCentric@24
D3DXVec2CatmullRom@24
D3DXVec2Hermite@24
D3DXVec2Normalize@8
D3DXVec2Transform@12
D3DXVec2TransformCoord@12
D3DXVec2TransformNormal@12
D3DXVec3BaryCentric@24
D3DXVec3CatmullRom@24
D3DXVec3Hermite@24
D3DXVec3Normalize@8
D3DXVec3Project@24
D3DXVec3Transform@12
D3DXVec3TransformCoord@12
D3DXVec3TransformNormal@12
D3DXVec3Unproject@24
D3DXVec4BaryCentric@24
D3DXVec4CatmullRom@24
D3DXVec4Cross@16
D3DXVec4Hermite@24
D3DXVec4Normalize@8
D3DXVec4Transform@12
D3DXWeldVertices@24
