//-------------------------------------------------------------------------
/*
Copyright (C) 1997, 2005 - 3D Realms Entertainment

This file is part of Shadow Warrior version 1.2

Shadow Warrior is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

Original Source: 1997 - Frank Maddin and Jim Norwood
Prepared for public release: 03/28/2005 - Charlie Wiederhold, 3D Realms
*/
//-------------------------------------------------------------------------

#define S_INAGAME   "YOU ARE IN A GAME!"
#define S_PRESSKEY  "PRESS ANY KEY WHEN READY"
#define S_MESSAGEON "GAME MESSAGES ARE NOW ON"
#define S_MESSAGEOFF "GAME MESSAGES ARE NOW OFF"

#define S_QUITYN "PRESS (Y) TO QUIT, (N) TO FIGHT ON."
#define S_QUIT1  "CONFUCIOUS SAY, 'HE WHO QUITS,"
#define S_QUIT1A "TOTALLY SUCKS!'"
#define S_QUIT2  "GO AHEAD, LEAVE, BUT WHILE YOU'RE"
#define S_QUIT2A "GONE, EVIL WILL BE TAKING OVER!"
#define S_QUIT3  "ALL OF JAPAN IS COUNTING ON YOU!"
#define S_QUIT3A "ONLY A LOSER COULD LET THEM DOWN."
#define S_QUIT4  "THIS IS TOO MUCH!  NOW WHERE"
#define S_QUIT4A "DID I PUT THOSE CHEAT CODES?"
#define S_QUIT5  "PRESS (Y) TO RUN HOME TO MOMMY,"
#define S_QUIT5A "(N) TO BE A MAN."
#define S_QUIT6  "YOU DO KNOW, THAT WHILE YOU'RE"
#define S_QUIT6A "AWAY, WE'LL BE RESTOCKING MT. FUJI"
#define S_QUIT6B "WITH A MILLION NEW MONSTERS?"
#define S_QUIT7  "COME BACK ANY TIME, WE ARE ALWAYS"
#define S_QUIT7A "LOOKING FOR A SACK OF RIPPER CHOW!"
#define S_QUIT8  "LET'S MAKE A STEALTHY EXIT TO DOS!"
