//-------------------------------------------------------------------------
/*
Copyright (C) 1997, 2005 - 3D Realms Entertainment

This file is part of Shadow Warrior version 1.2

Shadow Warrior is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

Original Source: 1997 - Frank Maddin and Jim Norwood
Prepared for public release: 03/28/2005 - Charlie Wiederhold, 3D Realms
*/
//-------------------------------------------------------------------------

//Be careful when changing this file - it is parsed by Editart and Build.
#define ALPHABET 85
#define MENU_MAIN 295
#define MENU_OPTION 296
#define MENU_WHICH 297
#define MENU_CURSOR 298
#define DART1 2000
#define DART2 2001
#define DART3 2002
#define DART4 2003
#define DART 2004
#define FIRE1 2005
#define FIRE2 2006
#define FIRE3 2007
#define FIRE4 2008
#define FIRE5 2009
#define FIREBALL1 2010
#define FIREBALL2 2011
#define FIREBALL3 2012
#define FIREBALL4 2013
#define FLAMES1 2014
#define FLAMES2 2015
#define FLAMES3 2016
#define FLAMES4 2017
#define FLAMES5 2018
#define GAS1 2019
#define GAS2 2020
#define GAS3 2021
#define GAS4 2022
#define GAS5 2023
#define GAS6 2024
#define GAS7 2025
#define GASCLOUD1 2026
#define GASCLOUD2 2027
#define GASCLOUD3 2028
#define GASCLOUD4 2029
#define GASCLOUD5 2030
#define GASCLOUD6 2031
#define GASCLOUD7 2032
#define GASCLOUD8 2033
#define GASCLOUD9 2034
#define LIGHT1 2035
#define LIGHT2 2036
#define LIGHT3 2037
#define LIGHT4 2038
#define LIGHT5 2039
#define LIGHT6 2040
#define ELECTRO1 2041
#define ELECTRO2 2042
#define ELECTRO3 2043
#define RIFLE1 2044
#define RIFLE2 2045
#define RIFLE3 2046
#define DONTUSE_EXTRA 2047
#define RIFLE4 2048
#define CARTRIGE3 2053
#define STATUE1 2054
#define STATUE2 2055
#define STATUE3 2056
#define STATUE4 2057
#define SWORD3 2065
#define SWORD4 2066
#define SWORD1 2067
#define SWORD2 2068
#define SWORD5 2069
#define BLACK___ 2306
#define FragBarErase 2375
#define FragBarErase2 2376
