// resource ids
#define WIN_STARTWIN        1000
#define WIN_STARTWINPAGE_CONFIG 2000
#define WIN_STARTWIN_BITMAP 100 // banner bitmap
#define WIN_STARTWIN_TABCTL 101
#define WIN_STARTWIN_CANCEL IDCANCEL
#define WIN_STARTWIN_START  IDOK

#define WIN_STARTWIN_MESSAGES   104 // output list box

#define RSRC_ICON       100
#define RSRC_BMP        200

// config page
#define IDCFULLSCREEN   100
#define IDC3DVMODE  101
#define IDCALWAYSSHOW   102
